"""
Overview:
    Meta information for treevalue package.

.. data:: __TITLE__

    Title of this project (should be `treevalue`).

.. data:: __VERSION__

    Version of this project.

.. data:: __AUTHOR__

    Author of this project.

.. data:: __AUTHOR_EMAIL__

    Email of the author.
"""

__TITLE__ = "treevalue"
__VERSION__ = "0.0.1"
__AUTHOR__ = "HansBug, DI-engine's Contributors"
__AUTHOR_EMAIL__ = "hansbug@buaa.edu.cn"
